use serde::Serialize;

use super::render_context::RenderContext;
use crate::error::CustomError;
use crate::intermediate::IObject;

use super::angle_link::RenderAngleLink;
use super::bold::RenderBold;
use super::citation::RenderCitation;
use super::citation_reference::RenderCitationReference;
use super::code::RenderCode;
use super::entity::RenderEntity;
use super::export_snippet::RenderExportSnippet;
use super::footnote_reference::RenderFootnoteReference;
use super::inline_babel_call::RenderInlineBabelCall;
use super::inline_source_block::RenderInlineSourceBlock;
use super::italic::RenderItalic;
use super::latex_fragment::RenderLatexFragment;
use super::line_break::RenderLineBreak;
use super::macros::render;
use super::org_macro::RenderOrgMacro;
use super::plain_link::RenderPlainLink;
use super::plain_text::RenderPlainText;
use super::radio_link::RenderRadioLink;
use super::radio_target::RenderRadioTarget;
use super::regular_link::RenderRegularLink;
use super::statistics_cookie::RenderStatisticsCookie;
use super::strike_through::RenderStrikeThrough;
use super::subscript::RenderSubscript;
use super::superscript::RenderSuperscript;
use super::target::RenderTarget;
use super::timestamp::RenderTimestamp;
use super::underline::RenderUnderline;
use super::verbatim::RenderVerbatim;

#[derive(Debug, Serialize)]
#[serde(untagged)]
pub(crate) enum RenderObject {
    Bold(RenderBold),
    Italic(RenderItalic),
    Underline(RenderUnderline),
    StrikeThrough(RenderStrikeThrough),
    Code(RenderCode),
    Verbatim(RenderVerbatim),
    PlainText(RenderPlainText),
    RegularLink(RenderRegularLink),
    RadioLink(RenderRadioLink),
    RadioTarget(RenderRadioTarget),
    PlainLink(RenderPlainLink),
    AngleLink(RenderAngleLink),
    OrgMacro(RenderOrgMacro),
    Entity(RenderEntity),
    LatexFragment(RenderLatexFragment),
    ExportSnippet(RenderExportSnippet),
    FootnoteReference(RenderFootnoteReference),
    Citation(RenderCitation),
    CitationReference(RenderCitationReference),
    InlineBabelCall(RenderInlineBabelCall),
    InlineSourceBlock(RenderInlineSourceBlock),
    LineBreak(RenderLineBreak),
    Target(RenderTarget),
    StatisticsCookie(RenderStatisticsCookie),
    Subscript(RenderSubscript),
    Superscript(RenderSuperscript),
    Timestamp(RenderTimestamp),
}

render!(RenderObject, IObject, original, render_context, {
    match original {
        IObject::Bold(inner) => Ok(RenderObject::Bold(RenderBold::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Italic(inner) => Ok(RenderObject::Italic(RenderItalic::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Underline(inner) => Ok(RenderObject::Underline(RenderUnderline::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::StrikeThrough(inner) => Ok(RenderObject::StrikeThrough(RenderStrikeThrough::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Code(inner) => Ok(RenderObject::Code(RenderCode::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Verbatim(inner) => Ok(RenderObject::Verbatim(RenderVerbatim::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::PlainText(inner) => Ok(RenderObject::PlainText(RenderPlainText::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::RegularLink(inner) => Ok(RenderObject::RegularLink(RenderRegularLink::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::RadioLink(inner) => Ok(RenderObject::RadioLink(RenderRadioLink::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::RadioTarget(inner) => Ok(RenderObject::RadioTarget(RenderRadioTarget::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::PlainLink(inner) => Ok(RenderObject::PlainLink(RenderPlainLink::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::AngleLink(inner) => Ok(RenderObject::AngleLink(RenderAngleLink::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::OrgMacro(inner) => Ok(RenderObject::OrgMacro(RenderOrgMacro::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Entity(inner) => Ok(RenderObject::Entity(RenderEntity::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::LatexFragment(inner) => Ok(RenderObject::LatexFragment(RenderLatexFragment::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::ExportSnippet(inner) => Ok(RenderObject::ExportSnippet(RenderExportSnippet::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::FootnoteReference(inner) => Ok(RenderObject::FootnoteReference(
            RenderFootnoteReference::new(render_context.clone(), inner)?,
        )),
        IObject::Citation(inner) => Ok(RenderObject::Citation(RenderCitation::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::CitationReference(inner) => Ok(RenderObject::CitationReference(
            RenderCitationReference::new(render_context.clone(), inner)?,
        )),
        IObject::InlineBabelCall(inner) => Ok(RenderObject::InlineBabelCall(
            RenderInlineBabelCall::new(render_context.clone(), inner)?,
        )),
        IObject::InlineSourceBlock(inner) => Ok(RenderObject::InlineSourceBlock(
            RenderInlineSourceBlock::new(render_context.clone(), inner)?,
        )),
        IObject::LineBreak(inner) => Ok(RenderObject::LineBreak(RenderLineBreak::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Target(inner) => Ok(RenderObject::Target(RenderTarget::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::StatisticsCookie(inner) => Ok(RenderObject::StatisticsCookie(
            RenderStatisticsCookie::new(render_context.clone(), inner)?,
        )),
        IObject::Subscript(inner) => Ok(RenderObject::Subscript(RenderSubscript::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Superscript(inner) => Ok(RenderObject::Superscript(RenderSuperscript::new(
            render_context.clone(),
            inner,
        )?)),
        IObject::Timestamp(inner) => Ok(RenderObject::Timestamp(RenderTimestamp::new(
            render_context.clone(),
            inner,
        )?)),
    }
});
